from __future__ import annotations

from pysam.libcbcf import VariantRecord, VariantRecordSample

default_filter = ["<NON_REF>"]


def get_alleles_str(variant: VariantRecord) -> str:
    """
    Get a string representing the possible alleles of a variant

    Parameters
    ----------
    variant : VariantRecord
        a pysam VariantRecord

    Returns
    -------
    comma joined alleles string,
        e.g: ref=A alt=G.*
        get_alleles_str(v) = 'A,G,*'
    """
    return ",".join(variant.alleles)


def get_filtered_alleles_list(variant: VariantRecord, filter_list: list[str] = None) -> list[str]:
    """
    Get a filtered list of alleles of a variant

    Parameters
    ----------
    variant : VariantRecord
        a pysam VariantRecord
    filter_list : list[str]
        a list of alleles to be filtered, initialized to default_filter (global var)

    Returns
    -------
    list of alleles, excluding the filter_list (initialized to default_filter), and * as minor allele
        e.g: ref=A alt=G,* filter_list=['*']
             get_alleles_str(v) = ['A','G']
    """
    if filter_list is None:
        filter_list = default_filter
    filtered_alleles = [a for a in variant.alleles if a not in filter_list]

    # discard minor allele '*'
    if filtered_alleles[-1] == "*":
        filtered_alleles = filtered_alleles[:-1]
    return filtered_alleles


def get_filtered_alleles_str(variant: VariantRecord, filter_list: list[str] = None) -> str:
    """
    Get a filtered list of alleles of a variant (as string)

    Parameters
    ----------
    variant : VariantRecord
        a pysam VariantRecord
    filter_list : list[str]
        a list of alleles to be filtered, initialized to default_filter (global var)

    Returns
    -------
    Comma joined alleles string, excluding the filter_list (initialized to default_filter), and * as minor allele
        e.g: ref=A alt=G,* filter_list=['*']
             get_alleles_str(v) = 'A,G'
    """
    return ",".join(get_filtered_alleles_list(variant, filter_list))


def get_genotype(variant_record_sample: VariantRecordSample) -> str:
    """
    Get genotype string for a specific sample in a specific variant.
    e.g: ref=A, alt=T, GT=0/1 -> 'A/T'

    Parameters
    ----------
    variant_record_sample: VariantRecordSample
        a pysam sample record of a variant (genotype, AD, GQ, etc)

    Returns
    -------
    genotype of sample
    """
    alleles = ["." if a is None else str(a) for a in variant_record_sample.alleles]
    return "/".join(alleles)


def get_genotype_indices(variant_record_sample: VariantRecordSample) -> str:
    """
    Get genotype indices string for a specific sample in a specific variant.
    e.g: ref=A, alt=T, GT=0/1 -> '0/1'

    Parameters
    ----------
    variant_record_sample: VariantRecordSample
        a pysam sample record of a variant (genotype, AD, GQ, etc)

    Returns
    -------
    genotype indices of sample as string
    """
    allele_indices = ["." if a is None else str(a) for a in variant_record_sample.allele_indices]
    return "/".join(sorted(allele_indices))


def has_candidate_alternatives(variant: VariantRecord) -> bool:
    """
    Returns
    -------
    True iff position has a candidate alternative allele (besides <NON_REF>)
    """
    candidate_alleles_str = get_filtered_alleles_str(variant)
    return "," in candidate_alleles_str


def is_snp(alleles: list[str]):
    """
    Return True if alleles represent a SNP
    - If the SNP locus is also within a deletion (* allele), still return True.
    - Ignored <NON_REF> allele

    Parameters
    ----------
    alleles : list[str]
        list of alleles

    Returns
    -------
    True iff the alleles represent SNP locus (all are of size 1)
    """
    __alleles = [a for a in alleles if a != "<NON_REF>"]
    return len(__alleles) >= 2 and all(len(a) == 1 for a in __alleles)  # noqa PLR2004


DNA_SYMBOLS = ["A", "T", "G", "C"]


def is_symbolic(allele: str) -> bool:
    """Check if allele is symbolic

    Parameters
    ----------
    allele : str
        Input allele

    Returns
    -------
    bool
        T/F
    """
    return len([x for x in allele if x not in DNA_SYMBOLS]) > 0


def is_indel(rec: VariantRecord) -> list[bool]:
    """Checks for each alt allele of the variant if it is indel

    Parameters
    ----------
    rec : pysam.VariantRecord
        Input variant

    Returns
    -------
    List[bool]
        list of T/F
    """
    return [len(x) != len(rec.alleles[0]) and not is_symbolic(x) for x in rec.alleles]


def is_deletion(rec: VariantRecord) -> list[bool]:
    """For every allele in the variant record - report if the allele is deletion

    Parameters
    ----------
    rec : pysam.VariantRecord
        Report if the variant is deletion for each allele

    Returns
    -------
    List[bool]
        List of T/F
    """
    return [len(x) < len(rec.alleles[0]) and not is_symbolic(x) for x in rec.alleles]


def is_insertion(rec: VariantRecord) -> list[bool]:
    """For every allele in the variant record - report if the allele is insertion

    Parameters
    ----------
    rec : pysam.VariantRecord
        Report if the variant is deletion for each allele

    Returns
    -------
    List[bool]
        List of T/F
    """
    return [len(x) > len(rec.alleles[0]) and not is_symbolic(x) for x in rec.alleles]


def indel_length(rec: VariantRecord) -> list[int]:
    """Length of all indel alleles

    Parameters
    ----------
    rec : pysam.VariantRecord
        Input record

    Returns
    -------
    List[int]
        List of length differences between the allele and the reference allele
    """
    return [abs(len(x) - len(rec.alleles[0])) if not is_symbolic(x) else 0 for x in rec.alleles]



import json
from argparse import ArgumentParser
from pathlib import Path

import nbformat
import papermill
from nbconvert import HTMLExporter
from ugbio_core.logger import logger


def modify_jupyter_notebook_html(
    input_html: str,
    output_html: str = None,
    font_size: int = 24,
    font_family: str = "Arial, sans-serif",
    max_width: str = "800px",
):
    """
    Modify the style of a Jupyter notebook HTML export.

    Parameters
    ----------
    input_html : str
        Path to the input HTML file.
    output_html : str, optional
        Path to the output HTML file. If not provided, the input HTML file will be modified in-place.
    font_size : int, optional
        The desired font size in pixels. Default is 16.
    font_family : str, optional
        The desired font family. Default is "Arial, sans-serif".
    max_width : str, optional
        The maximum width of the content. Default is "700px".

    """

    # Define the CSS to insert.
    css = f"""
    body {{
      font-size: {font_size}px;
      font-family: {font_family};
      margin: 0 auto;
      max-width: {max_width};
      text-align: left;
    }}
    div.output_text {{
      font-size: {font_size}px;
      font-family: {font_family};
      text-align: left;
    }}
    """

    # Read the HTML file.
    with open(input_html, encoding="utf-8") as file:
        html = file.read()

    # Insert the CSS into the HTML.
    html = html.replace("</head>", f'<style type="text/css">{css}</style></head>')

    # Write the updated HTML back to the file.
    output_html = output_html if output_html else input_html
    with open(output_html, "w", encoding="utf-8") as file:
        file.write(html)


def generate_report(
    template_notebook_path: Path, parameters: dict, output_report_html_path: Path, tmp_files: list[Path] = None
) -> None:  # type: ignore
    """
    Generate report based on jupyter notebook template.

    Parameters
    ----------
    template_notebook_path : Path
        Path to jupyter notebook template
    parameters : dict
        Parameters for report
    output_report_html_path : Path
        Path to output html report file
    tmp_files : list[Path]
        List of temporary files to be removed after report generation

    """
    if tmp_files is None:
        tmp_files = []

    output_report_ipynb = output_report_html_path.with_suffix(".ipynb")
    tmp_files.append(output_report_ipynb)

    # inject parameters and run notebook
    logger.info(f"Executing notebook (with papermill): {template_notebook_path}")

    def convert_path_to_str(value):
        if isinstance(value, Path):
            return str(value)
        elif isinstance(value, list):
            return [str(p) if isinstance(p, Path) else p for p in value]
        else:
            return value

    parameters = {k: convert_path_to_str(v) for k, v in parameters.items()}
    papermill.execute_notebook(
        input_path=str(template_notebook_path),
        output_path=str(output_report_ipynb),
        parameters=parameters,
        kernel_name="python3",
    )

    # convert to html
    logger.info(f"Converting notebook {output_report_ipynb} to HTML file {output_report_html_path}")
    notebook = nbformat.read(str(output_report_ipynb), as_version=4)
    html_exporter = HTMLExporter()
    html_exporter.exclude_input = True
    (body, resources) = html_exporter.from_notebook_node(notebook)

    with open(output_report_html_path, "w") as f:
        f.write(body)

    # edit html for readability
    logger.info(f"Modifying HTML style for {output_report_html_path}")
    modify_jupyter_notebook_html(output_report_html_path)

    # remove temporary files
    logger.debug(f"Removing tmp_files={[str(tmp_file) for tmp_file in tmp_files]}")
    for temp_file in tmp_files:
        if temp_file.is_file():
            temp_file.unlink()


def main():
    parser = ArgumentParser()
    parser.add_argument("--template_path", type=Path, help="Path to jupyter notebook template", required=True)
    parser.add_argument(
        "--params",
        type=json.loads,
        help='Parameters for the report, in a json format. E.g. \'{"a":"b", "c":"d"}\'',
        required=True,
    )
    parser.add_argument("--output_path", type=Path, help="Path to output html report file", required=True)
    parser.add_argument(
        "--tmp_files",
        type=str,
        help="Comma-separated list of temporary files to be removed after report generation",
        required=False,
    )
    args = parser.parse_args()

    # Split the tmp_files argument into a list of Path objects
    tmp_files = [Path(file) for file in args.tmp_files.split(",")] if args.tmp_files else []

    generate_report(args.template_path, args.params, args.output_path, tmp_files)


if __name__ == "__main__":
    main()



from __future__ import annotations

import pysam
from pysam import VariantRecord


class BufferedVariantReader:
    """
    A wrapper around pysam VariantFile.
    To be we used when random access to relatively sparse set of variants is needed.

    Reads a vcf/gvcf (can be zipped) files, and enables to efficiently fetch a specific variant from locus.
    pysam.VariantFile fetch returns a block of variants on each fetch.
    Here we cache this block of variants for quicker access in the future.
    """

    # pylint:disable=too-few-public-methods
    def __init__(self, file_name: str):
        self.pysam_reader = pysam.VariantFile(file_name)
        self.header = self.pysam_reader.header
        self.variants = {}
        self.current_chromosome = ""

    def get_variant(self, chromosome: str, pos: int) -> VariantRecord | None:
        if chromosome == self.current_chromosome:
            if pos in self.variants:
                return self.variants[pos]
        self.current_chromosome = chromosome
        self.variants = {}

        for variant in self.pysam_reader.fetch(chromosome, pos - 1, pos + 1000):
            self.variants[variant.pos] = variant

        if pos in self.variants:
            return self.variants[pos]
        return None



import os
from pathlib import Path

from simppl.simple_pipeline import SimplePipeline
from ugbio_core.exec_utils import print_and_execute
from ugbio_core.logger import logger


class IntervalFile:
    def __init__(
        self,
        sp: SimplePipeline = None,
        interval: str | None = None,
        ref: str | None = None,
        ref_dict: str | None = None,
    ):
        self.sp = sp
        self.files_to_clean = []
        # determine the file type and create the other temporary copy
        if interval is None:
            self._is_none: bool = True
            self._interval_list_file_name: str | None = None
            self._bed_file_name: str | None = None

        elif interval.endswith(".interval_list"):
            self._interval_list_file_name = interval
            # create the interval bed file
            if not os.path.exists(f"{os.path.splitext(interval)[0]}.bed"):
                self.__execute(f"picard IntervalListToBed I={interval} O={os.path.splitext(interval)[0]}.bed")
                self.files_to_clean.append(f"{os.path.splitext(interval)[0]}.bed")

            self._bed_file_name = f"{os.path.splitext(interval)[0]}.bed"
            self._is_none = False

        elif interval.endswith(".bed"):
            self._bed_file_name = interval
            # deduce ref_dict
            if ref_dict is None:
                ref_dict = f"{ref}.dict"
            if not os.path.isfile(ref_dict):
                logger.error(f"dict file does not exist: {ref_dict}")
                return
            # create the interval list file
            if not os.path.exists(f"{os.path.splitext(interval)[0]}.interval_list"):
                self.__execute(
                    f"picard BedToIntervalList I={interval} "
                    f"O={os.path.splitext(interval)[0]}.interval_list SD={ref_dict}"
                )
                self.files_to_clean.append(f"{os.path.splitext(interval)[0]}.interval_list")

            self._interval_list_file_name = f"{os.path.splitext(interval)[0]}.interval_list"
            self._is_none = False
        else:
            logger.error("the cmp_intervals should be of type interval list or bed")
            self._is_none = True
            self._interval_list_file_name = None
            self._bed_file_name = None

    def as_bed_file(self):
        return self._bed_file_name

    def as_interval_list_file(self):
        return self._interval_list_file_name

    def is_none(self):
        return self._is_none

    def __execute(self, command: str, output_file: str | None = None):
        print_and_execute(command, output_file=output_file, simple_pipeline=self.sp, module_name=__name__)

    def __del__(self):
        for file in self.files_to_clean:
            if Path(file).is_file():
                os.remove(file)



class Genotype:
    def __init__(self, genotype: str):
        """
        @param genotype - / separated genotype indices. e.g '0/0', '0/1', '0', '1'
        """
        self.genotype_str = genotype

    def convert_to_reference(self) -> str:
        return "/".join(["0" for _ in self.genotype_str.split("/")])

    def is_reference(self):
        return all(a == "0" for a in self.genotype_str.split("/"))


def sort_gt(gt: tuple[int] | str) -> tuple[int]:
    gt_list = []
    if isinstance(gt, tuple):
        gt_list = list(gt)
    elif isinstance(gt, str):
        gt_list = [int(a) for a in gt.replace("/", "|").split("|")]

    gt_list.sort(key=lambda e: (e is None, e))
    return tuple(gt_list)


def different_gt(gt1: tuple[int] | str, gt2: tuple[int] | str) -> bool:
    out = sort_gt(gt1) != sort_gt(gt2)
    return out



from __future__ import annotations

import numpy as np
import pandas as pd
from scipy.stats import multinomial
from sklearn import metrics
from ugbio_core.math_utils import safe_divide

# Goodness of fit functions


def scale_contingency_table(table: list[int], n: int) -> list[int]:
    """
    Parameters
    ----------
    table: List[int]
        contingency table represented as list of counts
    n: int
        scale-factor
    Returns
    -------
    A contingency table where the sum of categories is approximately n
    """
    sum_table = sum(table)
    if sum_table > 0:
        scaled_table = np.array(table) * (n / sum_table)
        return list(np.round(scaled_table).astype(int))
    # if sum_table == 0:
    return table


def correct_multinomial_frequencies(counts: list[int]) -> np.ndarray:
    """

    Parameters
    ----------
    counts - List[int]
        contingency table represented as list of ints

    Returns
    -------
    Frequency of each category after add-one correction
    """
    corrected_counts = np.array(counts) + 1
    return corrected_counts / np.sum(corrected_counts)


def multinomial_likelihood(actual: list[int], expected: list[int]) -> float:
    """
    Calculate the likelihood of observing actual, under the multinomial distribution described by expected

    Parameters
    ----------
    actual : List[int]
        a list of counts (integer observations)
    expected : List[int]
        a list counts from which to fit a multinomial distribution to
    Returns
    -------
    Likelihood of expected distribution given actual observation
    """
    freq_expected = correct_multinomial_frequencies(expected)
    return multinomial.pmf(x=actual, n=sum(actual), p=freq_expected)


def multinomial_likelihood_ratio(actual: list[int], expected: list[int]) -> tuple[float, float]:
    likelihood = multinomial_likelihood(actual, expected)
    max_likelihood = multinomial_likelihood(actual, actual)
    likelihood_ratio = likelihood / max_likelihood
    return likelihood, likelihood_ratio


# Metrics functions


def get_precision(false_positives: int, true_positives: int, return_if_denominator_is_0=1) -> float:
    """
    Get the precision, as defined from tp and false_positives

    Parameters
    ----------
    false_positives : int
        number of false positive observation (detected but false)
    true_positives : int
        number of true positive observations (detected true)
    return_if_denominator_is_0: any
        return value if tp + fp == 0
    Returns
    -------
    The precision score
    """
    if false_positives + true_positives == 0:
        return return_if_denominator_is_0
    return 1 - false_positives / (false_positives + true_positives)


def get_recall(false_negatives: int, true_positives: int, return_if_denominator_is_0=1) -> float:
    """
    Get the recall, as defined from true_positives and false_negatives

    Parameters
    ----------
    false_negatives : int
        number of false negative observations (missed true)
    true_positives : int
        number of true positive observation (detected true)
    return_if_denominator_is_0: any
        return value if tp + fn == 0

    Returns
    -------
    The recall score
    """
    if false_negatives + true_positives == 0:
        return return_if_denominator_is_0
    return 1 - false_negatives / (false_negatives + true_positives)


def get_f1(precision: float, recall, null_value=np.nan) -> float:
    """
    Get the F1 score (harmonic mean of precision and recall)

    Parameters
    ----------
    precision : float
        precision of the experiment
    recall : float
        recall of the experiment
    null_value: Any
        return this null_value if recall or precision are equal to this null_value

    Returns
    -------
    The F1 score of the experiment
    """
    if null_value in {precision, recall}:
        return null_value
    return safe_divide(2 * precision * recall, precision + recall)


def precision_recall_curve(
    gtr: np.ndarray | pd.Series,
    predictions: np.ndarray | pd.Series,
    fn_mask: np.ndarray | pd.Series,
    pos_label: str | int | None = 1,
    min_class_counts_to_output: int = 20,
) -> tuple:
    """Calculates precision/recall curve from double prediction scores and ground truth
    Similar to sklearn precision_recall curve, but uses mask of variants that were false negatives

    Parameters
    ----------
    gtr: np.ndarray
        String array of ground truth labels. FN should be labeled 'tp'
    predictions: np.ndarray
        Array of prediction scores
    fn_mask: np.ndarray
        Boolean array that has true on the locations that were false negatives
    pos_label: str or 0
        Label of true call, otherwise 1s will be considered
    min_class_counts_to_output: int
        Limit on the count of classes (denominator) below which the results are too noisy and not calculated

    Returns
    -------
    tuple
        precisions, recalls, f1_score, prediction_values
    """

    if len(gtr) == 0:
        return np.array([]), np.array([]), np.array([]), np.array([])

    assert len(set(gtr)) <= 2, "Only up to two classes of variant labels are possible"  # noqa S101 PLR2004
    assert len(fn_mask) == len(predictions), "FN mask should be of the length of predictions"  # noqa S101

    gtr_select = gtr[~fn_mask]
    gtr_select = gtr_select == pos_label
    predictions_select = predictions[~fn_mask]
    original_fn_count = fn_mask.sum()

    if len(gtr_select) > 0:
        raw_precision, raw_recall, thresholds = metrics.precision_recall_curve(
            gtr_select, predictions_select, pos_label=True
        )
    else:
        raw_precision = np.array([gtr_select.sum() / len(gtr_select), 1.0])
        raw_recall = np.array([1.0, 0.0])
        if len(predictions_select) == 0:
            thresholds = np.array([0])
        else:
            thresholds = np.array([np.min(predictions_select)])

    recall_correction = gtr_select.sum() / (gtr_select.sum() + original_fn_count)
    recalls = raw_recall * recall_correction
    recalls = recalls[
        1:-1
    ]  # remove the 1,0 value that sklearn adds, remove the initial value that the precision_recall_curve adds
    precisions = raw_precision[1:-1]
    thresholds = thresholds[1:]
    f1_score = 2 * (recalls * precisions) / (recalls + precisions + np.finfo(float).eps)

    # Find the score cutoff at which too few calls remain (to remove areas where the precision recall curve is noisy)
    predictions_select = np.sort(predictions_select)
    if len(predictions_select) > 0:
        threshold_cutoff = predictions_select[max(0, len(predictions_select) - min_class_counts_to_output)]
    else:
        threshold_cutoff = 0

    mask = thresholds > threshold_cutoff
    return precisions[~mask], recalls[~mask], f1_score[~mask], thresholds[~mask]


def generate_sample_from_dist(vals: np.ndarray, probs: np.ndarray) -> np.ndarray:
    """Returns values from a distribution

    Parameters
    ----------
    vals: np.ndarray
        Values
    probs: np.ndarray
        Probabilities

    Returns
    -------
    np.ndarray
    """
    return np.random.Generator.choice(vals, 10000, p=probs)



from enum import Enum


class StrandDirection(Enum):
    """
    A representation of strand direction
    """

    FORWARD = 1
    REVERSE = 0
    UNKNOWN = -1


def is_forward_strand(allele: str):
    """
    pileup alleles with uppercase letters represent forward strand
    todo - handle indels
    """
    return any(x in allele for x in ("A", "C", "G", "T"))
